"Other resource dictionaries"

drivers = {
    "postgresql": "postgresql+psycopg2",
    "mysql": "mysql+pymysql",
    "sqlite": "sqlite",
    "oracle": "oracle+oracledb",
    "mssql": "mssql+pymssql",
    "mariadb": "mariadb+pymysql",
}

url_keys = (
    "drivername",
    "username",
    "password",
    "host",
    "port",
    "database",
    "query",
)
